"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const url = require("url");
function buildLD(schemabase) {
    function _label(jsd) {
        return path.parse(new url.URL(jsd.title).pathname).name;
    }
    function datatypeLD(jsd) {
        return {
            '@type': 'rdfs:Datatype',
            '@id': `sdo:${_label(jsd)}`,
            'rdfs:label': _label(jsd),
            'rdfs:comment': jsd.description,
        };
    }
    function classLD(jsd, propertybase) {
        function superclass(jsd) {
            return (_label(jsd) !== 'Thing') ? path.parse(jsd.allOf[0].$ref).name : null;
        }
        return {
            '@type': 'rdfs:Class',
            '@id': `sdo:${_label(jsd)}`,
            'rdfs:label': _label(jsd),
            'rdfs:comment': jsd.description,
            'rdfs:subClassOf': (superclass(jsd)) ? { '@id': `sdo:${superclass(jsd)}` } : null,
            'rdfs:member': Object.entries(jsd.allOf[1].properties).map((entry) => {
                let prop_name = entry[0];
                let memberjsd = propertybase.find((sch) => sch.title === `http://schema.org/${prop_name}`) || null;
                if (memberjsd)
                    return { '@id': `sdo:${prop_name}` };
                else
                    throw new ReferenceError(`No corresponding jsd file was found for member subschema \`${_label(jsd)}#${prop_name}\`.`);
            }),
            superClassOf: [],
            valueOf: [],
        };
    }
    function propertyLD(jsd) {
        function superproperty(jsd) {
            return (jsd.allOf[0] !== true) ? path.parse(jsd.allOf[0].$ref).name.split('.')[0] : null;
        }
        return {
            '@type': 'rdf:Property',
            '@id': `sdo:${_label(jsd)}`,
            'rdfs:label': _label(jsd),
            'rdfs:comment': jsd.description,
            'rdfs:subPropertyOf': (jsd.allOf[0] !== true) ? { '@id': `sdo:${superproperty(jsd)}` } : null,
            'rdfs:domain': [],
            'rdfs:range': jsd.definitions.ExpectedType.anyOf.map((schema) => ({ '@id': `sdo:${path.parse(schema.$ref).name}` })),
            superPropertyOf: [],
            $rangeArray: jsd.allOf[1].anyOf.length === 2,
        };
    }
    function processSubclasses(jsonld, classbase) {
        let superclass = jsonld['rdfs:subClassOf'];
        let referenced = (superclass) ? classbase.find((c) => c['@id'] === superclass['@id']) || null : null;
        if (referenced) {
            referenced.superClassOf.push({ '@id': jsonld['@id'] });
        }
    }
    function processSubproperties(jsonld, propertybase) {
        let superproperty = jsonld['rdfs:subPropertyOf'];
        let referenced = (superproperty) ? propertybase.find((p) => p['@id'] === superproperty['@id']) || null : null;
        if (referenced) {
            referenced.superPropertyOf.push({ '@id': jsonld['@id'] });
        }
    }
    function processDomains(jsonld, propertybase) {
        jsonld['rdfs:member'].forEach((property_ref) => {
            let referenced = propertybase.find((p) => p['@id'] === property_ref['@id']) || null;
            if (referenced) {
                referenced['rdfs:domain'].push({ '@id': jsonld['@id'] });
            }
        });
    }
    function processValueOf(jsonld, classbase) {
        jsonld['rdfs:range'].forEach((class_ref) => {
            let referenced = classbase.find((c) => c['@id'] === class_ref['@id']) || null;
            if (referenced) {
                referenced.valueOf.push({ '@id': jsonld['@id'] });
            }
        });
    }
    const datatypeJSDs = schemabase.filter((jsd) => jsd.$schema === 'http://json-schema.org/draft-07/schema#');
    const classJSDs = schemabase.filter((jsd) => jsd.$schema === 'https://chharvey.github.io/schemaorg-jsd/meta/type.jsd#');
    const propertyJSDs = schemabase.filter((jsd) => jsd.$schema === 'https://chharvey.github.io/schemaorg-jsd/meta/member.jsd#');
    const datatypeLDs = datatypeJSDs.map((jsd) => datatypeLD(jsd));
    const classLDs = classJSDs.map((jsd) => classLD(jsd, propertyJSDs));
    const propertyLDs = propertyJSDs.map((jsd) => propertyLD(jsd));
    classLDs.forEach((jsonld) => {
        processSubclasses(jsonld, classLDs);
        processDomains(jsonld, propertyLDs);
    });
    propertyLDs.forEach((jsonld) => {
        processSubproperties(jsonld, propertyLDs);
        processValueOf(jsonld, classLDs);
    });
    return {
        "@context": {
            "sdo": "http://schema.org/",
            "rdf": "https://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "superClassOf": { "@reverse": "rdfs:subClassOf" },
            "superPropertyOf": { "@reverse": "rdfs:subPropertyOf" },
            "valueOf": { "@reverse": "rdfs:range" },
        },
        "@graph": [
            ...datatypeLDs,
            ...classLDs,
            ...propertyLDs,
        ],
    };
}
exports.buildLD = buildLD;
function buildTS(jsonlddocument) {
    function _linklist(lds) {
        return lds.map((obj) => ` * - {@link ${obj['@id'].split(':')[1]}}`).join('\n');
    }
    function datatypeTS(ld) {
        let SDODatatypeNames;
        (function (SDODatatypeNames) {
            SDODatatypeNames[SDODatatypeNames["Boolean"] = 0] = "Boolean";
            SDODatatypeNames[SDODatatypeNames["Date"] = 1] = "Date";
            SDODatatypeNames[SDODatatypeNames["DateTime"] = 2] = "DateTime";
            SDODatatypeNames[SDODatatypeNames["Integer"] = 3] = "Integer";
            SDODatatypeNames[SDODatatypeNames["Number"] = 4] = "Number";
            SDODatatypeNames[SDODatatypeNames["Text"] = 5] = "Text";
            SDODatatypeNames[SDODatatypeNames["Time"] = 6] = "Time";
            SDODatatypeNames[SDODatatypeNames["URL"] = 7] = "URL";
        })(SDODatatypeNames || (SDODatatypeNames = {}));
        let alias = ld['rdfs:label'];
        let type = ({
            Boolean: 'boolean',
            Date: 'string',
            DateTime: 'string',
            Integer: 'number',
            Number: 'number',
            Text: 'string',
            Time: 'string',
            URL: 'string',
        })[alias];
        return `
			/**
			 * ${ld['rdfs:comment']}
			 *
			 * @see http://schema.org/${ld['rdfs:label']}
			 */
			export type ${alias} = ${type}
		`;
    }
    function classTS(ld) {
        return `
			/**
			 * ${ld['rdfs:comment']}
			 *
			 * ${(ld['superClassOf'].length) ? `*(Non-Normative):* Known subclasses:\n${_linklist(ld['superClassOf'])}\n` : ''}
			 * ${(ld['valueOf'].length) ? `*(Non-Normative):* May appear as values of:\n${_linklist(ld['valueOf']).replace(/}/g, '_type}')}\n` : ''}
			 * @see http://schema.org/${ld['rdfs:label']}
			 */
			export interface ${ld['rdfs:label']} extends ${(ld['rdfs:subClassOf']) ? ld['rdfs:subClassOf']['@id'].split(':')[1] : 'JSONLDObject'} {
				${ld['rdfs:member'].map((member) => member['@id'].split(':')[1]).map((name) => `
					${name}?: ${name}_type
				`).join('')}
			}
		`;
    }
    function propertyTS(ld) {
        let rangeunion = `${ld['rdfs:range'].map((cls) => cls['@id'].split(':')[1]).join('|')}`;
        return `
			/**
			 * ${ld['rdfs:comment']}
			 *
			 * ${(ld['rdfs:subPropertyOf']) ? `Extends {@link ${ld['rdfs:subPropertyOf']['@id'].split(':')[1]}}` : ''}
			 * ${(ld['superPropertyOf'].length) ? `*(Non-Normative):* Known subproperties:\n${_linklist(ld['superPropertyOf'])}\n` : ''}
			 * ${(ld['rdfs:domain'].length) ? `*(Non-Normative):* Property of:\n${_linklist(ld['rdfs:domain'])}\n` : ''}
			 * @see http://schema.org/${ld['rdfs:label']}
			 */
			type ${ld['rdfs:label']}_type = ${rangeunion}${(ld['$rangeArray']) ? `|(${rangeunion})[]` : ''}
		`;
    }
    const JSONLD = jsonlddocument['@graph'];
    return [
        `import { JSONLDObject } from '@chharvey/requirejson'`,
        ...JSONLD.filter((jsonld) => jsonld['@type'] === 'rdfs:Datatype').map((ld) => datatypeTS(ld)),
        ...JSONLD.filter((jsonld) => jsonld['@type'] === 'rdfs:Class').map((ld) => classTS(ld)),
        ...JSONLD.filter((jsonld) => jsonld['@type'] === 'rdf:Property').map((ld) => propertyTS(ld)),
    ].join('');
}
exports.buildTS = buildTS;
